/*
 * Copyright © 2022
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <NetworkManager.h>
#include <adwaita.h>

G_BEGIN_DECLS

typedef struct _CcQrCodeDialog CcQrCodeDialog;

#define CC_TYPE_QR_CODE_DIALOG (cc_qr_code_dialog_get_type ())
G_DECLARE_FINAL_TYPE (CcQrCodeDialog, cc_qr_code_dialog, CC, QR_CODE_DIALOG, AdwDialog);
GtkWidget *cc_qr_code_dialog_new (NMConnection *connection);

G_END_DECLS
